@@ -11,8 +11,10 @@
 use core::iter::Sum;
 use core::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Rem, RemAssign, Sub, SubAssign};
 
+use crate::UnitTypes;
 use crate::{
-    traits::{Lerp, SimdVec, SimdVecFloat},
+    bindings::{Vector, VectorFloat},
+    traits::Lerp,
     Point2, Point3, Point4, Scalar, Size2, Size3, Unit,
 };
 
@@ -114,15 +116,6 @@ crate::impl_scaling!(Vector4, 4 [f32, f64, i32, u32]);
 macro_rules! impl_vector {
     ($base_type_name:ident [ $dimensions:literal ] => $vec_ty:ident, $point_ty:ident $(, $size_ty:ident)?) => {
         impl<T: Unit> $base_type_name<T> {
-            #[doc = "Instantiate vector with all components set to 2."]
-            #[doc = ""]
-            #[doc = "This is a shorthand for `Self::one() + Self::one()`."]
-            #[inline]
-            #[must_use]
-            pub fn two() -> Self {
-                Self::one() + Self::one()
-            }
-
             #[doc = "Multiply all components of this vector with a scalar value."]
             #[doc = ""]
             #[doc = "This exists as a method because `Mul<T::Scalar>` cannot be implemented"]
@@ -255,8 +248,8 @@ macro_rules! impl_vector {
 
         impl<T> $base_type_name<T>
         where
-            T: crate::traits::UnitTypes,
-            T::$vec_ty: SimdVecFloat<$dimensions, Scalar = T::Primitive>,
+            T: UnitTypes,
+            T::$vec_ty: VectorFloat<$dimensions, Scalar = T::Primitive>,
         {
             #[doc = "Normalize the vector."]
             #[doc = ""]
@@ -379,13 +372,13 @@ macro_rules! impl_vector {
         impl<T: Unit> Sum for $base_type_name<T> {
             #[must_use]
             fn sum<I: Iterator<Item = Self>>(iter: I) -> Self {
-                iter.fold(Self::zero(), Add::add)
+                iter.fold(Self::ZERO, Add::add)
             }
         }
 
         impl<T> Lerp<T::Primitive> for $base_type_name<T>
         where
-            T: crate::traits::UnitTypes,
+            T: UnitTypes,
             T::$vec_ty: Lerp<T::Primitive>,
         {
             #[inline]
@@ -403,24 +396,19 @@ impl_vector!(Vector4 [4] => Vec4, Point4);
 
 impl<T: Unit> Vector2<T> {
     /// Unit vector in the direction of the X axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_x() -> Self {
-        Self {
-            x: T::Scalar::one(),
-            y: T::Scalar::zero(),
-        }
-    }
+    pub const X: Self = Vector2 {
+        x: T::Scalar::ONE,
+        y: T::Scalar::ZERO,
+    };
 
     /// Unit vector in the direction of the Y axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_y() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::one(),
-        }
-    }
+    pub const Y: Self = Vector2 {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ONE,
+    };
+
+    /// The unit axes.
+    pub const AXES: [Self; 2] = [Self::X, Self::Y];
 
     /// Select components of this vector and return a new vector containing
     /// those components.
@@ -443,37 +431,26 @@ impl<T: Unit> Vector2<T> {
 
 impl<T: Unit> Vector3<T> {
     /// Unit vector in the direction of the X axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_x() -> Self {
-        Self {
-            x: T::Scalar::one(),
-            y: T::Scalar::zero(),
-            z: T::Scalar::zero(),
-        }
-    }
-
+    pub const X: Self = Self {
+        x: T::Scalar::ONE,
+        y: T::Scalar::ZERO,
+        z: T::Scalar::ZERO,
+    };
     /// Unit vector in the direction of the Y axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_y() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::one(),
-            z: T::Scalar::zero(),
-        }
-    }
-
+    pub const Y: Self = Self {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ONE,
+        z: T::Scalar::ZERO,
+    };
     /// Unit vector in the direction of the Z axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_z() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::zero(),
-            z: T::Scalar::one(),
-        }
-    }
+    pub const Z: Self = Self {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ZERO,
+        z: T::Scalar::ONE,
+    };
+
+    /// The unit axes.
+    pub const AXES: [Self; 3] = [Self::X, Self::Y, Self::Z];
 
     /// Select components of this vector and return a new vector containing
     /// those components.
@@ -543,61 +520,36 @@ where
 
 impl<T: Unit> Vector4<T> {
     /// Unit vector in the direction of the X axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_x() -> Self {
-        Self {
-            x: T::Scalar::one(),
-            y: T::Scalar::zero(),
-            z: T::Scalar::zero(),
-            w: T::Scalar::zero(),
-        }
-    }
-
+    pub const X: Self = Self {
+        x: T::Scalar::ONE,
+        y: T::Scalar::ZERO,
+        z: T::Scalar::ZERO,
+        w: T::Scalar::ZERO,
+    };
     /// Unit vector in the direction of the Y axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_y() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::one(),
-            z: T::Scalar::zero(),
-            w: T::Scalar::zero(),
-        }
-    }
-
+    pub const Y: Self = Self {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ONE,
+        z: T::Scalar::ZERO,
+        w: T::Scalar::ZERO,
+    };
     /// Unit vector in the direction of the Z axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_z() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::zero(),
-            z: T::Scalar::one(),
-            w: T::Scalar::zero(),
-        }
-    }
-
+    pub const Z: Self = Self {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ZERO,
+        z: T::Scalar::ONE,
+        w: T::Scalar::ZERO,
+    };
     /// Unit vector in the direction of the W axis.
-    #[inline]
-    #[must_use]
-    pub fn unit_w() -> Self {
-        Self {
-            x: T::Scalar::zero(),
-            y: T::Scalar::zero(),
-            z: T::Scalar::zero(),
-            w: T::Scalar::one(),
-        }
-    }
+    pub const W: Self = Self {
+        x: T::Scalar::ZERO,
+        y: T::Scalar::ZERO,
+        z: T::Scalar::ZERO,
+        w: T::Scalar::ONE,
+    };
 
-    /// Horizontal maximum, taking `w` into account as well.
-    ///
-    /// Workaround for [https://github.com/bitshifter/glam-rs/issues/283](https://github.com/bitshifter/glam-rs/issues/283).
-    #[inline]
-    #[must_use]
-    pub fn max_element_w(self) -> T::Scalar {
-        self.x.max(self.y.max(self.z.max(self.w)))
-    }
+    /// The unit axes.
+    pub const AXES: [Self; 4] = [Self::X, Self::Y, Self::Z, Self::W];
 
     /// Select components of this vector and return a new vector containing
     /// those components.
@@ -612,10 +564,28 @@ crate::impl_mint!(Vector2, 2, Vector2);
 crate::impl_mint!(Vector3, 3, Vector3);
 crate::impl_mint!(Vector4, 4, Vector4);
 
+impl<T: UnitTypes<Vec3 = glam::Vec3>> Mul<Vector3<T>> for glam::Quat {
+    type Output = Vector3<T>;
+
+    fn mul(self, rhs: Vector3<T>) -> Self::Output {
+        Vector3::from_raw(self * rhs.to_raw())
+    }
+}
+
+impl<T: UnitTypes<Vec3 = glam::DVec3>> Mul<Vector3<T>> for glam::DQuat {
+    type Output = Vector3<T>;
+
+    fn mul(self, rhs: Vector3<T>) -> Self::Output {
+        Vector3::from_raw(self * rhs.to_raw())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use approx::assert_abs_diff_eq;
 
+    use crate::vector;
+
     use super::*;
 
     struct F32;
@@ -756,6 +726,28 @@ mod tests {
         check_splat_ints!(10, add(2), 12);
         check_splat_ints!(2, sub(1), 1);
         check_splat_ints!(10, sub(2), 8);
+
+        let mut v: Vector4<f32> = vector!(1.0, 2.0, 3.0, 4.0);
+        v *= Vector4::splat(2.0);
+        assert_eq!(v, (2.0, 4.0, 6.0, 8.0));
+        v /= Vector4::splat(2.0);
+        assert_eq!(v, (1.0, 2.0, 3.0, 4.0));
+        v += Vector4::splat(1.0);
+        assert_eq!(v, (2.0, 3.0, 4.0, 5.0));
+        v -= Vector4::splat(2.0);
+        assert_eq!(v, (0.0, 1.0, 2.0, 3.0));
+        v %= Vector4::splat(2.0);
+        assert_eq!(v, (0.0, 1.0, 0.0, 1.0));
+    }
+
+    #[test]
+    fn sum() {
+        let a: Vector4<f32> = vector!(1.0, 2.0, 3.0, 4.0);
+        let b: Vector4<f32> = vector!(1.0, 2.0, 3.0, 4.0);
+        let c: Vector4<f32> = vector!(1.0, 2.0, 3.0, 4.0);
+        let d: Vector4<f32> = vector!(1.0, 2.0, 3.0, 4.0);
+        let sum: Vector4<f32> = [a, b, c, d].into_iter().sum();
+        assert_eq!(sum, (4.0, 8.0, 12.0, 16.0));
     }
 
     #[test]
@@ -779,15 +771,16 @@ mod tests {
 
     #[test]
     fn units() {
-        assert_abs_diff_eq!(Vec2::unit_x(), (1.0, 0.0));
-        assert_abs_diff_eq!(Vec2::unit_y(), (0.0, 1.0));
-        assert_abs_diff_eq!(Vec3::unit_x(), (1.0, 0.0, 0.0));
-        assert_abs_diff_eq!(Vec3::unit_y(), (0.0, 1.0, 0.0));
-        assert_abs_diff_eq!(Vec3::unit_z(), (0.0, 0.0, 1.0));
-        assert_abs_diff_eq!(Vec4::unit_x(), (1.0, 0.0, 0.0, 0.0));
-        assert_abs_diff_eq!(Vec4::unit_y(), (0.0, 1.0, 0.0, 0.0));
-        assert_abs_diff_eq!(Vec4::unit_z(), (0.0, 0.0, 1.0, 0.0));
-        assert_abs_diff_eq!(Vec4::unit_w(), (0.0, 0.0, 0.0, 1.0));
+        use crate::vector;
+        assert_abs_diff_eq!(Vec2::X, vector!(1.0, 0.0));
+        assert_abs_diff_eq!(Vec2::Y, vector!(0.0, 1.0));
+        assert_abs_diff_eq!(Vec3::X, vector!(1.0, 0.0, 0.0));
+        assert_abs_diff_eq!(Vec3::Y, vector!(0.0, 1.0, 0.0));
+        assert_abs_diff_eq!(Vec3::Z, vector!(0.0, 0.0, 1.0));
+        assert_abs_diff_eq!(Vec4::X, vector!(1.0, 0.0, 0.0, 0.0));
+        assert_abs_diff_eq!(Vec4::Y, vector!(0.0, 1.0, 0.0, 0.0));
+        assert_abs_diff_eq!(Vec4::Z, vector!(0.0, 0.0, 1.0, 0.0));
+        assert_abs_diff_eq!(Vec4::W, vector!(0.0, 0.0, 0.0, 1.0));
     }
 
     #[test]
@@ -815,13 +808,13 @@ mod tests {
         assert!(Vec4::nan().is_nan());
 
         // Replace NaNs with zeroes.
-        let v = Vec4::select(v4.is_nan_mask(), Vec4::zero(), v4);
+        let v = Vec4::select(v4.is_nan_mask(), Vec4::ZERO, v4);
         assert_eq!(v, (1.0, 2.0, 0.0, 4.0));
     }
 
     #[test]
     fn swizzle2() {
-        assert_eq!(Vec2::unit_x().swizzle::<1, 0>(), (0.0, 1.0));
+        assert_eq!(Vec2::X.swizzle::<1, 0>(), (0.0, 1.0));
         assert_eq!(
             Vec3::new(1.0, 2.0, 3.0).swizzle2::<1, 0>(),
             Vec2::new(2.0, 1.0)
@@ -834,10 +827,7 @@ mod tests {
             Vec3::new(1.0, 2.0, 3.0).swizzle::<2, 1, 0>(),
             (3.0, 2.0, 1.0)
         );
-        assert_eq!(
-            Vec2::unit_x().swizzle3::<1, 0, 1>(),
-            Vec3::new(0.0, 1.0, 0.0)
-        );
+        assert_eq!(Vec2::X.swizzle3::<1, 0, 1>(), Vec3::new(0.0, 1.0, 0.0));
         assert_eq!(
             Vec4::new(1.0, 2.0, 3.0, 4.0).swizzle3::<3, 2, 1>(),
             Vec3::new(4.0, 3.0, 2.0)
@@ -852,7 +842,7 @@ mod tests {
         );
 
         assert_eq!(
-            Vec2::unit_x().swizzle4::<1, 0, 1, 0>(),
+            Vec2::X.swizzle4::<1, 0, 1, 0>(),
             Vec4::new(0.0, 1.0, 0.0, 1.0)
         );
         assert_eq!(
@@ -863,12 +853,13 @@ mod tests {
 
     #[test]
     fn to_3d() {
-        assert_eq!(Vec2::unit_x().to_3d(2.0), Vec3::new(1.0, 0.0, 2.0));
+        let v = Vec2 { x: 3.0, y: 4.0 };
+        assert_eq!(v.to_3d(2.0), crate::vec3!(3.0, 4.0, 2.0));
     }
 
     #[test]
     fn to_4d() {
-        assert_eq!(Vec3::unit_z().to_4d(2.0), Vec4::new(0.0, 0.0, 1.0, 2.0));
+        assert_eq!(Vec3::Z.to_4d(2.0), Vec4::new(0.0, 0.0, 1.0, 2.0));
     }
 
     #[test]
@@ -888,11 +879,23 @@ mod tests {
         {
             let x: Vec4 = (1.0, 2.0, 3.0, 4.0).into();
 
-            let a = x * 2.0;
-            let b = x / 2.0;
+            let a = x.mul_scalar(2.0);
+            let b = x.div_scalar(2.0);
+            let c = x.rem_scalar(2.0);
 
             assert_eq!(a, (2.0, 4.0, 6.0, 8.0));
             assert_eq!(b, (0.5, 1.0, 1.5, 2.0));
+            assert_eq!(c, (1.0, 0.0, 1.0, 0.0));
+
+            let mut a = x;
+            let mut b = x;
+            let mut c = x;
+            a *= 2.0;
+            b /= 2.0;
+            c %= 2.0;
+            assert_eq!(a, (2.0, 4.0, 6.0, 8.0));
+            assert_eq!(b, (0.5, 1.0, 1.5, 2.0));
+            assert_eq!(c, (1.0, 0.0, 1.0, 0.0));
         }
         {
             let x: DVec4 = (1.0, 2.0, 3.0, 4.0).into();
@@ -945,9 +948,7 @@ mod tests {
         assert_eq!(a.min(b), [1.0, 2.0, 1.0, 3.0]);
         assert_eq!(a.max(b), [4.0, 2.0, 3.0, 4.0]);
         assert_eq!(a.min_element(), 1.0);
-        // See https://github.com/bitshifter/glam-rs/issues/283
-        // assert_eq!(a.max_element(), 3.0);
-        assert_eq!(a.max_element_w(), 4.0);
+        assert_eq!(a.max_element(), 4.0);
     }
 
     #[test]
@@ -977,6 +978,12 @@ mod tests {
         assert_abs_diff_eq!(a.normalize(), b);
     }
 
+    #[test]
+    fn length() {
+        let v = Vec3::new(1.0, 2.0, 3.0);
+        assert_eq!(v.length(), glam::Vec3::new(1.0, 2.0, 3.0).length());
+    }
+
     #[test]
     fn exp() {
         let a = Vec3::splat(1.0);
@@ -1015,84 +1022,26 @@ mod tests {
         assert_eq!(a.signum(), (1.0, -1.0, -1.0));
     }
 
-    #[cfg(feature = "std")]
     #[test]
-    fn debug_type_name() {
-        extern crate alloc;
+    fn rotate() {
+        use crate::Angle;
+        use approx::assert_abs_diff_eq;
 
-        let untyped_f32: Vector2<f32> = Vector2 { x: 123.0, y: 456.0 };
-        let untyped_f64: Vector2<f64> = Vector2 { x: 123.0, y: 456.0 };
-        let untyped_u32: Vector2<u32> = Vector2 { x: 123, y: 456 };
+        let v = Vector3::<f32>::X;
+        let quat = Angle::from_degrees(180.0f32).to_rotation(Vector3::Z);
+        assert_abs_diff_eq!(quat * v, -v);
 
-        assert_eq!(
-            alloc::format!("{:?}", untyped_f32),
-            "Vector2<f32> { x: 123.0, y: 456.0 }"
-        );
-        assert_eq!(
-            alloc::format!("{:?}", untyped_f64),
-            "Vector2<f64> { x: 123.0, y: 456.0 }"
-        );
-        assert_eq!(
-            alloc::format!("{:?}", untyped_u32),
-            "Vector2<u32> { x: 123, y: 456 }"
-        );
+        let v = Vector3::<f64>::X;
+        let quat = Angle::from_degrees(180.0f64).to_rotation(Vector3::Z);
+        assert_abs_diff_eq!(quat * v, -v);
+    }
 
-        let untyped_i32: Vector2<i32> = Vector2 { x: 123, y: 456 };
-        assert_eq!(
-            alloc::format!("{:?}", untyped_i32),
-            "Vector2<i32> { x: 123, y: 456 }"
-        );
-        assert_eq!(
-            alloc::format!("{:#?}", untyped_i32),
-            r#"
-Vector2<i32> {
-    x: 123,
-    y: 456,
-}"#
-            .trim()
-        );
-
-        enum UnnamedUnit {}
-        impl Unit for UnnamedUnit {
-            type Scalar = i32;
-        }
-
-        let unnamed: Vector2<UnnamedUnit> = Vector2 { x: 123, y: 456 };
-        assert_eq!(
-            alloc::format!("{:?}", unnamed),
-            "Vector2 { x: 123, y: 456 }"
-        );
-        assert_eq!(
-            alloc::format!("{:#?}", unnamed),
-            r#"
-Vector2 {
-    x: 123,
-    y: 456,
-}"#
-            .trim()
-        );
-
-        enum CustomName {}
-        impl Unit for CustomName {
-            type Scalar = i32;
-            fn name() -> Option<&'static str> {
-                Some("Custom")
-            }
-        }
-
-        let custom: Vector2<CustomName> = Vector2 { x: 123, y: 456 };
-        assert_eq!(
-            alloc::format!("{:?}", custom),
-            "Vector2<Custom> { x: 123, y: 456 }"
-        );
-        assert_eq!(
-            alloc::format!("{:#?}", custom),
-            r#"
-Vector2<Custom> {
-    x: 123,
-    y: 456,
-}"#
-            .trim()
-        );
+    #[test]
+    fn matrix_mul_custom_unit() {
+        use crate::{vec3, Matrix3};
+        let mat = Matrix3::<f32>::IDENTITY;
+        let a: Vector3<F32> = vec3!(20.0, 30.0, 1.0);
+        let b: Vector3<F32> = mat * a;
+        assert_eq!(b, (20.0, 30.0, 1.0));
     }
 }
